from enum import Enum
from functools import partial

import jax
import jax.numpy as jnp


class PVCPatchPadding(Enum):
    ZERO = 1
    BLEED = 2


def precompute_bitreverse_indices(n):
    def _alg(i):
        m = i * 2
        return jnp.concatenate([m, m + 1])

    o = jnp.array([0])
    while len(o) < n:
        o = _alg(o)
    return o


def precompute_cfkt_constants(n):
    c = [jnp.array([1])]
    while len(c) < n:
        c = [
            jnp.concatenate([y, y * jnp.exp(1j * jnp.pi * i / len(c))])
            for i, y in enumerate(c + c)
        ]
    c = jnp.array(c)
    return c


def precompute_rfkt_constants(n):
    theta = jnp.pi / n
    first_half = jnp.arange((n >> 1) + 1)
    ws = jnp.exp(1j * first_half * theta)
    return ws


def make_pvc_patches(i, hop_length, window_length, padding=PVCPatchPadding.BLEED):
    assert len(i.shape) == 3
    batch_size = i.shape[0]
    o = jax.lax.conv_general_dilated_patches(
        jnp.transpose(i, (0, 2, 1)),
        filter_shape=(window_length,),
        window_strides=(hop_length,),
        padding=(
            ((window_length - hop_length, window_length - hop_length),)
            if padding == PVCPatchPadding.BLEED
            else ((0, 0),)
        ),
    )
    return jnp.transpose(
        jnp.reshape(o, (batch_size, window_length, -1)),
        (0, 2, 1),
    )


def cfkt(xc, bitreverse_indices, c):
    nd = len(xc) * 2
    xc = xc[bitreverse_indices]
    xc = jnp.dot(c, xc)
    # Scale output
    scale = 1.0 / nd
    xc *= scale
    return xc


PI = jnp.pi
TWOPI = 2.0 * jnp.pi


def rfkt(xx, bitreverse_indices, c, ws):
    n = len(xx) // 2
    x = xx[::2] + 1j * xx[1::2]
    x = cfkt(x, bitreverse_indices, c)
    c1 = 0.5
    c2 = -c1
    x = jnp.concatenate([x, jnp.array([x[0]])])
    first_half = jnp.arange((n >> 1) + 1)
    second_half = -first_half - 1  # Mirrored indices
    x1 = x[first_half]
    x3 = x[second_half]
    h1 = c1 * (x1 + jnp.conjugate(x3))
    h2 = c2 * (jnp.conjugate(x1) - x3)
    h2 = h2.imag + 1j * h2.real
    x = jnp.concatenate(
        [(h1 + ws * h2)[:-1], jnp.flip(jnp.conjugate(h1) - jnp.conjugate(ws * h2))]
    )
    x = jnp.concatenate([jnp.array([x[0].real + 1j * x[-1].real]), x[1:]])
    x = x[:-1]
    real = jnp.real(x)
    imag = jnp.imag(x)
    x_final = jnp.ravel(jnp.column_stack((real, imag)))

    return x_final


def koonce_sinc(i, n, nw):
    x = (-(nw - 1) / 2.0) + jnp.arange(nw)
    pix = jnp.pi * x
    return i * n * jnp.sin(pix / n) / pix


def koonce_normalization(x):
    return x * 2.0 / jnp.sum(x)


def adjust_values(n, n_bins):
    adjustment = jnp.where(n < 0, ((-n - 1) // n_bins + 1) * n_bins, 0)
    n_adjusted = n + adjustment
    n_final = n_adjusted % n_bins
    return n_final


def fold_pvc_patches(x, w, n_bins, hop_length, window_length):
    assert len(x.shape) == 3
    batch_size = x.shape[0]
    seq = x.shape[1]
    seq_range = jnp.arange(seq) * hop_length - (window_length - hop_length)
    seq_range = jnp.expand_dims(
        window_length - adjust_values(seq_range, n_bins), axis=(0, -1)
    )
    m = x * jnp.expand_dims(w, axis=(0, 1))
    win_range = jnp.arange(window_length)
    m = jnp.concatenate([m, m], axis=-1)
    m = jax.vmap(lambda x, y: x[:, jnp.squeeze(y) + win_range], in_axes=1, out_axes=1)(
        m, seq_range
    )
    padding = 0 if window_length % n_bins == 0 else n_bins - (window_length % n_bins)
    padding_config = [(0, 0) for _ in range(m.ndim - 1)] + [(0, padding)]
    m = jnp.pad(m, padding_config)
    m = jnp.reshape(m, (batch_size, seq, -1, n_bins))
    m = jnp.sum(m, axis=2)
    return m


TWOPI = 2 * jnp.pi


def convert_dft_to_amp_and_freq(real, imag, lastphase, d, r):
    n_bins = real.shape[-1]
    fft_len = n_bins * 2

    # Compute constants
    fundamental = r / fft_len
    factor = r / (d * TWOPI)

    # Compute magnitude (amplitude)
    amplitude = jnp.hypot(real, imag)

    # Compute phase with negation as in the original C code: -atan2(imag, real)
    phase = -jnp.arctan2(imag, real)

    # Compute phase difference as in the C code
    phasediff = phase - lastphase

    # Phase unwrapping, ensuring it matches the logic in the C code
    phasediff = (phasediff + jnp.pi) % TWOPI - jnp.pi

    # Frequency calculation
    frequency = (
        phasediff * factor
        + jnp.expand_dims(jnp.arange(n_bins), axis=(0,)) * fundamental
    )

    # Update lastphase for next call, using the computed phase
    lastphase_updated = phase

    return lastphase_updated, jnp.stack([amplitude, frequency], axis=-1)


def convert_stft_to_amp_and_freq(real, imag, lastphase, d, r):
    stacked_ri = jnp.stack([real, imag], axis=-1)
    _, o = jax.lax.scan(
        lambda c, x: convert_dft_to_amp_and_freq(x[..., 0], x[..., 1], c, d, r),
        lastphase,
        jnp.transpose(stacked_ri, (1, 0, 2, 3)),
    )
    o = jnp.transpose(o, (1, 0, 2, 3))
    return o[..., 0], o[..., 1]


def convert_stft_to_amp_and_freq_using_0_phase(real, imag, d, r):
    return convert_stft_to_amp_and_freq(real, imag, jnp.zeros_like(real[:, 0, :]), d, r)


def noscbank_cell(lvix, chans, nw, p_inc, i_inv, rg):
    """
    Generates a bank of oscillators with varying amplitudes and frequencies over a specified range,
    accumulating their outputs. This function is designed to process and synthesize audio signals
    based on a set of input control parameters for amplitude and frequency modulation.

    Parameters:
    - lvix (tuple): A tuple containing two elements:
        - lv (array): The last value of amplitude and frequency for each oscillator.
        - ix (array): The current index or phase of each oscillator.
    - C (array): A 2D array of control parameters where each row contains an amplitude and frequency
      value for an oscillator. The array shape should be (-1, 2), where each row represents [amplitude, frequency].
    - Nw (int): The number of samples in the window, used for scaling the oscillator outputs.
    - Pinc (float): The phase increment per sample, derived from the sampling rate.
    - Iinv (float): The inverse of the interpolation factor, used for controlling the rate of change in amplitude and frequency.
    - rg (array): An array of indices used for generating the oscillator outputs over time.

    Returns:
    - A tuple containing:
        - A tuple with the last values of amplitude and frequency for each oscillator, and the final index or phase.
        - The accumulated output of all oscillators, summed over the second axis.

    This function utilizes JAX for efficient numerical computations and parallel processing of oscillators.
    The internal mechanism involves mapping a function over each oscillator's control parameters, followed
    by a loop to generate the oscillator's output using cosine modulation based on the current phase and frequency.
    The outputs are then summed to produce the final signal.

    Example usage:
    ```
    # Assuming lvix, C, Nw, Pinc, Iinv, and rg are defined as per the function requirements.
    last_values, index, output_signal = noscbank_cell(lvix, C, Nw, Pinc, Iinv, rg)
    ```
    """
    lv, ix = lvix

    def _vmap(c, lastval, index):

        finc = (c[1] * p_inc - lastval[1]) * i_inv
        ainc = (c[0] - lastval[0]) * i_inv
        address = index

        def _loop(carry, _):
            addr, lastval = carry
            idx = addr
            o_chan = lastval[0] * nw * jnp.cos(TWOPI * idx)
            addr += lastval[1]

            lastval += jnp.array([ainc, finc])
            return (addr, lastval), o_chan

        carry, o = jax.lax.scan(_loop, (address, lastval), rg)
        address = carry[0]
        lastval = jnp.array([c[0], c[1] * p_inc])
        index = address
        return o, lastval, index

    chans = jnp.reshape(chans, (-1, 2))
    o, lastval, index = jax.vmap(_vmap, in_axes=0, out_axes=0)(chans, lv, ix)
    return (lastval, index), jnp.sum(o, axis=0)


def noscbank(lvix, chans, nw, p_inc, i_inv, rg):
    return jax.lax.scan(
        partial(noscbank_cell, nw=nw, p_inc=p_inc, i_inv=i_inv, rg=rg), lvix, chans
    )



import jax.numpy as jnp
import scipy.signal
from jax import lax


def create_fir_filter(filter_type, coef, fs, ntaps):
    if ntaps % 2 == 0:
        raise ValueError(f"ntaps must be odd (ntaps={ntaps}).")

    if filter_type == "hp":
        taps = jnp.array([1, -coef, 0])
    elif filter_type == "fd":
        taps = jnp.array([1, 0, -coef])
    elif filter_type == "aw":
        f1, f2, f3, f4 = 20.598997, 107.65265, 737.86223, 12194.217
        a1000 = 1.9997
        nums = [(2 * jnp.pi * f4) ** 2 * (10 ** (a1000 / 20)), 0, 0, 0, 0]
        dens = jnp.polymul(
            jnp.array([1, 4 * jnp.pi * f4, (2 * jnp.pi * f4) ** 2]),
            jnp.array([1, 4 * jnp.pi * f1, (2 * jnp.pi * f1) ** 2]),
        )
        dens = jnp.polymul(
            jnp.polymul(dens, jnp.array([1, 2 * jnp.pi * f3])),
            jnp.array([1, 2 * jnp.pi * f2]),
        )
        b, a = scipy.signal.bilinear(nums, dens, fs=fs)
        w_iir, h_iir = scipy.signal.freqz(b, a, worN=512, fs=fs)
        taps = scipy.signal.firls(ntaps, w_iir, abs(h_iir), fs=fs)
    else:
        raise ValueError(f"Invalid filter type: {filter_type}")

    return taps


def fir_filter(signal, kernel, ntaps):
    # Reshape filter kernel for convolution
    padding = ntaps // 2
    kernel = kernel.reshape((1, 1, -1))

    # Apply FIR filter using 1D convolution
    filtered_signal = lax.conv_general_dilated(
        signal[:, None, :],  # Add channel dimension
        kernel,
        window_strides=(1,),
        padding=((padding, padding),),
        dimension_numbers=("NCH", "IOH", "NCH"),
        feature_group_count=1,
    )

    return filtered_signal.squeeze()  # Remove channel dimension



from functools import partial

import jax
import jax.numpy as jnp
import optax

from . import stft


def spectral_convergence_loss(x_mag, y_mag):
    """
    Calculate the spectral convergence loss.

    Args:
        x_mag (array): The magnitude spectrum of the first signal.
        y_mag (array): The magnitude spectrum of the second signal.

    Returns:
        The spectral convergence loss.
    """
    numerator = jnp.linalg.norm(y_mag - x_mag, ord=2)
    denominator = jnp.linalg.norm(y_mag, ord=2)
    loss = numerator / denominator
    return loss


def stft_magnitude_loss(x_mag, y_mag, log=True, distance="L1", reduction="mean"):
    """
    Calculate the STFT magnitude loss.

    Args:
        x_mag (array): The magnitude spectrum of the first signal.
        y_mag (array): The magnitude spectrum of the second signal.
        log (bool): Whether to log-scale the STFT magnitudes.
        distance (str): Distance function ["L1", "L2"].
        reduction (str): Reduction of the loss elements ["mean", "sum", "none"].

    Returns:
        The STFT magnitude loss.
    """
    if log:
        x_mag = jnp.log(x_mag + 1e-8)  # Adding a small value to avoid log(0)
        y_mag = jnp.log(y_mag + 1e-8)

    if distance == "L1":
        # L1 Loss (Mean Absolute Error)
        loss = jnp.abs(x_mag - y_mag)
    elif distance == "L2":
        # L2 Loss (Mean Squared Error)
        loss = (x_mag - y_mag) ** 2
    else:
        raise ValueError(f"Invalid distance: '{distance}'.")

    if reduction == "mean":
        return jnp.mean(loss)
    elif reduction == "sum":
        return jnp.sum(loss)
    elif reduction == "none":
        return loss
    else:
        raise ValueError(f"Invalid reduction: '{reduction}'.")


def stft_loss(
    traced_params,
    untraced_params,
    inputs,
    target,
    w_sc=1.0,
    w_log_mag=1.0,
    w_lin_mag=0.0,
    w_phs=0.0,
    scale=None,
    perceptual_weighting=None,
    eps=1e-8,
    output="loss",
    reduction="mean",
    mag_distance="L1",
):
    """
    Calculate the STFT loss.
    """
    bs, _, chs = inputs.shape

    # Compute STFT for inputs and target
    def _to_map(x):
        o = stft.stft(traced_params, untraced_params, x)
        return o

    to_map = jax.vmap(_to_map, in_axes=-1, out_axes=-1)

    def mf(x, y):
        return (
            jnp.sqrt(jnp.clip((x**2) + (y**2), a_min=eps)),
            jax.lax.atan2(y, x),
        )

    inputs_mag, inputs_phs = mf(*to_map(inputs))
    target_mag, target_phs = mf(*to_map(target))

    # Apply scaling (e.g., Mel, Chroma) if required
    if scale is not None:
        inputs_mag = jnp.matmul(scale, inputs_mag)
        target_mag = jnp.matmul(scale, target_mag)

    # Apply perceptual weighting if required
    if perceptual_weighting is not None:
        # since FIRFilter only support mono audio we will move channels to batch dim
        inputs = jnp.transpose(inputs, (0, 2, 1))
        target = jnp.transpose(target, (0, 2, 1))
        inputs = jnp.reshape(inputs, (bs * chs, -1))
        target = jnp.reshape(target, (bs * chs, -1))

        # now apply the filter to both
        inputs, target = perceptual_weighting(inputs), perceptual_weighting(target)

        # now move the channels back
        inputs = jnp.reshape(inputs, (bs, chs, -1))
        target = jnp.reshape(target, (bs, chs, -1))
        inputs = jnp.transpose(inputs, (0, 2, 1))
        target = jnp.transpose(target, (0, 2, 1))

    # Calculate loss components
    inputs_mag, inputs_phs, target_mag, target_phs = (
        jnp.ravel(inputs_mag),
        jnp.ravel(inputs_phs),
        jnp.ravel(target_mag),
        jnp.ravel(target_phs),
    )
    sc_mag_loss = (
        spectral_convergence_loss(inputs_mag, target_mag) * w_sc if w_sc else 0.0
    )
    log_mag_loss = (
        stft_magnitude_loss(
            inputs_mag, target_mag, log=True, reduction=reduction, distance=mag_distance
        )
        * w_log_mag
        if w_log_mag
        else 0.0
    )
    lin_mag_loss = (
        stft_magnitude_loss(
            inputs_mag,
            target_mag,
            log=False,
            reduction=reduction,
            distance=mag_distance,
        )
        * w_lin_mag
        if w_lin_mag
        else 0.0
    )
    phs_loss = (
        optax.squared_error(inputs_phs, target_phs).mean() * w_phs if w_phs else 0.0
    )

    # Combine loss components
    total_loss = sc_mag_loss + log_mag_loss + lin_mag_loss + phs_loss

    # Apply reduction (mean, sum)
    if reduction == "mean":
        total_loss = jnp.mean(total_loss)
    elif reduction == "sum":
        total_loss = jnp.sum(total_loss)

    # Return based on the output type
    if output == "loss":
        return total_loss
    elif output == "full":
        return total_loss, sc_mag_loss, log_mag_loss, lin_mag_loss, phs_loss


def multi_resolution_stft_loss(
    traced_params,
    untraced_params,
    inputs,
    target,
    w_sc=1.0,
    w_log_mag=1.0,
    w_lin_mag=0.0,
    w_phs=0.0,
    scale=None,
    perceptual_weighting=None,
    eps=1e-8,
    output="loss",
    reduction="mean",
    mag_distance="L1",
):
    mrstft_loss = 0.0
    sc_mag_loss, log_mag_loss, lin_mag_loss, phs_loss = [], [], [], []
    loss_fn = partial(
        stft_loss,
        inputs=inputs,
        target=target,
        w_sc=w_sc,
        w_log_mag=w_log_mag,
        w_lin_mag=w_lin_mag,
        w_phs=w_phs,
        scale=scale,
        perceptual_weighting=perceptual_weighting,
        eps=eps,
        output=output,
        reduction=reduction,
        mag_distance=mag_distance,
    )

    for p, q in zip(traced_params, untraced_params):
        if output == "full":
            tmp_loss = loss_fn(traced_params=p, untraced_params=q)
            mrstft_loss += tmp_loss[0]
            sc_mag_loss.append(tmp_loss[1])
            log_mag_loss.append(tmp_loss[2])
            lin_mag_loss.append(tmp_loss[3])
            phs_loss.append(tmp_loss[4])
        else:
            mrstft_loss += loss_fn(traced_params=p, untraced_params=q)

    mrstft_loss /= len(traced_params)

    if output == "loss":
        return mrstft_loss
    else:
        return mrstft_loss, sc_mag_loss, log_mag_loss, lin_mag_loss, phs_loss



import jax.numpy as jnp


def log_cosh_loss(inputs, target, a=1.0, eps=1e-8):
    losses = jnp.mean((1 / a) * jnp.log(jnp.cosh(a * (inputs - target)) + eps), axis=-2)
    losses = jnp.mean(losses)
    return losses


def esr_loss(inputs, target, eps=1e-8):
    num = jnp.sum(((target - inputs) ** 2), axis=1)
    denom = jnp.sum(target**2, axis=1) + eps
    losses = num / denom
    losses = jnp.mean(losses)
    return losses


def dc_loss(inputs, target, eps=1e-8):
    num = jnp.mean(target - inputs, axis=1) ** 2
    denom = jnp.mean(target**2, axis=1) + eps
    losses = num / denom
    losses = jnp.mean(losses)
    return losses



import dataclasses
from typing import Optional

import jax.lax as lax
import jax.numpy as jnp
import librosa


@dataclasses.dataclass(frozen=True)
class DftParams:
    n: int
    norm: Optional[str]


@dataclasses.dataclass(frozen=True)
class StftParams:
    hop_length: int
    n_fft: int


def init_dft_params(n, norm=None):
    """
    Initialize the DFT and IDFT matrices (parameters) for JAX operations.

    Args:
    n (int): Size of the DFT and IDFT matrices.
    norm (str or None): Normalization mode, either None or 'ortho'.

    Returns:
    A dictionary containing DFT and IDFT matrices as well as other parameters.
    """

    def dft_matrix(n):
        x, y = jnp.meshgrid(jnp.arange(n), jnp.arange(n))
        omega = jnp.exp(-2 * jnp.pi * 1j / n)
        w = omega ** (x * y)
        return w

    def idft_matrix(n):
        x, y = jnp.meshgrid(jnp.arange(n), jnp.arange(n))
        omega = jnp.exp(2 * jnp.pi * 1j / n)
        w = omega ** (x * y)
        return w

    w = dft_matrix(n)
    inv_w = idft_matrix(n)

    return {
        "W_real": jnp.real(w),
        "W_imag": jnp.imag(w),
        "inv_W_real": jnp.real(inv_w),
        "inv_W_imag": jnp.imag(inv_w),
    }, DftParams(n, norm)


def dft(traced_params, untraced_params, x_real, x_imag):
    """
    Perform the Discrete Fourier Transform (DFT) using JAX.

    Args:
    params (dict): Parameters containing the DFT matrices.
    x_real (array): Real part of the signal.
    x_imag (array): Imaginary part of the signal.

    Returns:
    Tuple of arrays representing the real and imaginary parts of the DFT.
    """
    z_real = jnp.dot(x_real, traced_params["W_real"]) - jnp.dot(
        x_imag, traced_params["W_imag"]
    )
    z_imag = jnp.dot(x_imag, traced_params["W_real"]) + jnp.dot(
        x_real, traced_params["W_imag"]
    )

    if untraced_params.norm == "ortho":
        z_real /= jnp.sqrt(untraced_params.n)
        z_imag /= jnp.sqrt(untraced_params.n)

    return z_real, z_imag


def idft(traced_params, untraced_params, x_real, x_imag):
    """
    Perform the Inverse Discrete Fourier Transform (IDFT) using JAX.

    Args:
    params (dict): Parameters containing the IDFT matrices.
    x_real (array): Real part of the signal.
    x_imag (array): Imaginary part of the signal.

    Returns:
    Tuple of arrays representing the real and imaginary parts of the IDFT.
    """
    z_real = jnp.dot(x_real, traced_params["inv_W_real"]) - jnp.dot(
        x_imag, traced_params["inv_W_imag"]
    )
    z_imag = jnp.dot(x_imag, traced_params["inv_W_real"]) + jnp.dot(
        x_real, traced_params["inv_W_imag"]
    )

    if untraced_params.norm is None:
        z_real /= untraced_params.n
    elif untraced_params.norm == "ortho":
        z_real /= jnp.sqrt(untraced_params.n)
        z_imag /= jnp.sqrt(untraced_params.n)

    return z_real, z_imag


def rdft(traced_params, untraced_params, x_real):
    """
    Perform the Right-side Real Discrete Fourier Transform (RDFT) using JAX.

    Args:
    traced_params (dict): Traced parameters containing the DFT matrices.
    untraced_params (dict): Untraced static parameters.
    x_real (array): Real part of the signal.

    Returns:
    Tuple of arrays representing the real and imaginary parts of the RDFT.
    """
    n_rfft = untraced_params.n // 2 + 1
    z_real = jnp.dot(x_real, traced_params["W_real"][..., :n_rfft])
    z_imag = jnp.dot(x_real, traced_params["W_imag"][..., :n_rfft])

    if untraced_params.norm == "ortho":
        z_real /= jnp.sqrt(untraced_params.n)
        z_imag /= jnp.sqrt(untraced_params.n)

    return z_real, z_imag


def irdft(traced_params, untraced_params, x_real, x_imag):
    """
    Perform the Inverse Real Discrete Fourier Transform (IRDFT) using JAX.

    Args:
    traced_params (dict): Traced parameters containing the IDFT matrices.
    untraced_params (dict): Untraced static parameters.
    x_real (array): Real part of the signal (n // 2 + 1,).
    x_imag (array): Imaginary part of the signal (n // 2 + 1,).

    Returns:
    An array representing the real part of the output signal.
    """
    n_rfft = untraced_params.n // 2 + 1

    # Flip and concatenate to reconstruct full signal
    flip_x_real = jnp.flip(x_real, axis=-1)
    flip_x_imag = jnp.flip(x_imag, axis=-1)

    x_real_full = jnp.concatenate((x_real, flip_x_real[..., 1 : n_rfft - 1]), axis=-1)
    x_imag_full = jnp.concatenate(
        (x_imag, -1.0 * flip_x_imag[..., 1 : n_rfft - 1]), axis=-1
    )

    # Calculate IRDFT
    z_real = jnp.dot(x_real_full, traced_params["inv_W_real"]) - jnp.dot(
        x_imag_full, traced_params["inv_W_imag"]
    )

    if untraced_params.norm == "ortho":
        z_real /= jnp.sqrt(untraced_params.n)

    return z_real


# from librosa
def pad_center(data, *, size, axis: int = -1):
    n = data.shape[axis]

    lpad = int((size - n) // 2)

    lengths = [(0, 0)] * data.ndim
    lengths[axis] = (lpad, int(size - n - lpad))

    if lpad < 0:
        raise ValueError(f"Target size ({size:d}) must be at least input size ({n:d})")

    return jnp.pad(data, lengths)


def init_stft_params(n_fft, hop_length=None, win_length=None, window_type="hann"):
    if win_length is None:
        win_length = n_fft

    if hop_length is None:
        hop_length = win_length // 4

    # Window function
    fft_window = jnp.array(librosa.filters.get_window(window_type, win_length))
    fft_window = pad_center(fft_window, size=n_fft)

    # DFT matrix
    def dft_matrix(n):
        x, y = jnp.meshgrid(jnp.arange(n), jnp.arange(n))
        omega = jnp.exp(-2 * jnp.pi * 1j / n)
        w = omega ** (x * y)
        return w

    w = dft_matrix(
        n_fft
    )  # Reuse the dft_matrix function from the previous implementation

    # Prepare convolutional filters
    real_filter = jnp.real(w[:, : n_fft // 2 + 1] * fft_window[:, None]).T
    imag_filter = jnp.imag(w[:, : n_fft // 2 + 1] * fft_window[:, None]).T

    return {
        "real_filter": real_filter[None, :],
        "imag_filter": imag_filter[None, :],
    }, StftParams(hop_length, n_fft)


def stft(traced_params, untraced_params, input_signal, center=True, pad_mode="reflect"):
    batch_size = input_signal.shape[0]
    n_fft = untraced_params.n_fft
    hop_length = untraced_params.hop_length

    if center:
        input_signal = jnp.pad(
            input_signal, ((0, 0), (n_fft // 2, n_fft // 2)), mode=pad_mode
        )

    # Convolution parameters
    stride = (hop_length,)

    # Real and Imaginary convolutions

    real = lax.conv_general_dilated(
        input_signal[:, None, :],
        traced_params["real_filter"],
        window_strides=stride,
        padding="VALID",
        lhs_dilation=None,
        rhs_dilation=None,
        dimension_numbers=("NCH", "IOH", "NCH"),
        feature_group_count=1,
    )

    imag = lax.conv_general_dilated(
        input_signal[:, None, :],
        traced_params["imag_filter"],
        window_strides=stride,
        padding="VALID",
        lhs_dilation=None,
        rhs_dilation=None,
        dimension_numbers=("NCH", "IOH", "NCH"),
        feature_group_count=1,
    )

    def last_reshape(x):
        x = jnp.reshape(x, (-1, batch_size, x.shape[1], x.shape[2]))
        x = jnp.sum(x, axis=0)
        x = jnp.transpose(x, (0, 2, 1))
        return x

    real, imag = last_reshape(real), last_reshape(imag)
    return real, imag
